# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import random
import itertools

import networkx as nx

import penaltymodel.core as pm

import dimod


class TestPenaltyModel(unittest.TestCase):
    def test_construction(self):

        # build a specification
        graph = nx.complete_graph(10)
        decision_variables = (0, 4, 5)
        feasible_configurations = {(-1, -1, -1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        # build a model
        model = dimod.BinaryQuadraticModel({v: 0 for v in graph},
                                           {edge: 0 for edge in graph.edges},
                                           0.0,
                                           vartype=dimod.SPIN)

        # build a PenaltyModel explicitly
        pm0 = pm.PenaltyModel(graph, decision_variables, feasible_configurations, dimod.SPIN, model, .1, 0)

        # build from spec
        pm1 = pm.PenaltyModel.from_specification(spec, model, .1, 0)

        # should result in equality
        self.assertEqual(pm0, pm1)

    def test_relabel(self):
        graph = nx.path_graph(3)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)
        linear = {v: 0 for v in graph}
        quadratic = {edge: -1 for edge in graph.edges}
        model = dimod.BinaryQuadraticModel(linear, quadratic, 0.0, vartype=dimod.SPIN)
        widget = pm.PenaltyModel.from_specification(spec, model, 2., -2)

        # now set up the same widget with 0 relabelled to 'a'
        graph = nx.path_graph(3)
        graph = nx.relabel_nodes(graph, {0: 'a'})
        decision_variables = ('a', 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)
        linear = {v: 0 for v in graph}
        quadratic = {edge: -1 for edge in graph.edges}
        model = dimod.BinaryQuadraticModel(linear, quadratic, 0.0, vartype=dimod.SPIN)
        test_widget = pm.PenaltyModel.from_specification(spec, model, 2., -2)

        # without copy
        new_widget = widget.relabel_variables({0: 'a'}, inplace=False)
        self.assertEqual(test_widget, new_widget)
        self.assertEqual(new_widget.decision_variables, ('a', 2))

        widget.relabel_variables({0: 'a'}, inplace=True)
        self.assertEqual(widget, test_widget)
        self.assertEqual(widget.decision_variables, ('a', 2))

    def test_bad_energy_range(self):
        graph = nx.path_graph(3)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        linear = {v: -3 for v in graph}
        quadratic = {edge: -1 for edge in graph.edges}
        model = dimod.BinaryQuadraticModel(linear, quadratic, 0.0, vartype=dimod.SPIN)
        with self.assertRaises(ValueError):
            widget = pm.PenaltyModel.from_specification(spec, model, 2., -2)

        linear = {v: 0 for v in graph}
        quadratic = {edge: 5 for edge in graph.edges}
        model = dimod.BinaryQuadraticModel(linear, quadratic, 0.0, vartype=dimod.SPIN)
        with self.assertRaises(ValueError):
            widget = pm.PenaltyModel.from_specification(spec, model, 2., -2)

    def test_relabel_forwards_and_backwards(self):
        graph = nx.path_graph(4)
        graph.add_edge(0, 2)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)
        linear = {v: 0 for v in graph}
        quadratic = {edge: -1 for edge in graph.edges}
        model = dimod.BinaryQuadraticModel(linear, quadratic, 0.0, vartype=dimod.SPIN)
        widget = pm.PenaltyModel.from_specification(spec, model, 2., -2)

        # make another one
        graph = nx.path_graph(4)
        graph.add_edge(0, 2)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)
        linear = {v: 0 for v in graph}
        quadratic = {edge: -1 for edge in graph.edges}
        model = dimod.BinaryQuadraticModel(linear, quadratic, 0.0, vartype=dimod.SPIN)
        original_widget = pm.PenaltyModel.from_specification(spec, model, 2., -2)

        new_label_sets = [(10, 1),
                          ('a', 'b'),
                          (1, 'b'),
                          ('1', '2', '3', '4'),
                          ('a', 'b', 'c', 'd')]
        new_label_sets.extend(itertools.permutations(graph))

        for new in new_label_sets:
            mapping = dict(enumerate(new))
            inv_mapping = {u: v for v, u in mapping.items()}

            # apply then invert with copy=False
            widget.relabel_variables(mapping, inplace=True)
            widget.relabel_variables(inv_mapping, inplace=True)
            self.assertEqual(widget, original_widget)

            # apply then invert with copy=True
            copy_widget = widget.relabel_variables(mapping, inplace=False)
            inv_copy = copy_widget.relabel_variables(inv_mapping, inplace=False)
            self.assertEqual(inv_copy, original_widget)

    def test_bqm_and_graph_label_matching(self):
        vartype = dimod.BINARY
        bqm = dimod.BinaryQuadraticModel({'a': -1, 'b': 0}, {('c', 'a'): 0}, 0,
                                         vartype)
        g1 = nx.complete_graph(['a', 1, 2])
        g2 = nx.complete_graph(['a', 'b', 'c'])

        with self.assertRaises(ValueError):
            pm.PenaltyModel(g1, ['a'], {(0, )}, vartype, bqm, 2, 0)

        pm.PenaltyModel(g2, ['a'], {(0, )}, vartype, bqm, 2, 0)



# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import networkx as nx
import penaltymodel


class TestAsGraph(unittest.TestCase):
    def test_int(self):
        self.assertEqual(penaltymodel.as_graph(0).nodes, nx.Graph().nodes)

        K10 = penaltymodel.as_graph(10)
        self.assertEqual(K10.nodes, nx.complete_graph(10).nodes)
        self.assertEqual(K10.edges, nx.complete_graph(10).edges)

    def test_sequence(self):
        self.assertEqual(penaltymodel.as_graph('').nodes, nx.Graph().nodes)

        K10 = penaltymodel.as_graph('abc')
        self.assertEqual(K10.nodes, nx.complete_graph('abc').nodes)
        self.assertEqual(K10.edges, nx.complete_graph('abc').edges)

    def test_graph(self):
        K0 = nx.Graph()
        self.assertIs(K0, penaltymodel.as_graph(K0))

        P6 = nx.path_graph(6)
        self.assertIs(P6, penaltymodel.as_graph(P6))



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import time
import unittest
import unittest.mock

import dimod
import networkx as nx

import penaltymodel

from penaltymodel.core import Specification, PenaltyModel
from penaltymodel.core import iter_caches, iter_factories, get_penalty_model
from penaltymodel.database import isolated_cache


class TestInterfaceWithCache(unittest.TestCase):
    @isolated_cache()
    def test_retrieval(self):
        # put some stuff in the database

        spec = Specification(nx.path_graph(2), (0, 1), {(-1, -1), (1, 1)}, vartype='SPIN')
        model = dimod.BinaryQuadraticModel({0: 0, 1: 0}, {(0, 1): -1}, 0.0, vartype='SPIN')
        widget = PenaltyModel.from_specification(spec, model, 2, -1)

        with self.assertWarns(DeprecationWarning):
            cache, = iter_caches()

        cache(widget)

        # now try to get it back
        with self.assertWarns(DeprecationWarning):
            new_widget = get_penalty_model(spec)

        self.assertEqual(widget, new_widget)

    @isolated_cache()
    def test_generation(self):
        # put some stuff in the database

        spec = Specification(nx.path_graph(2), (0, 1), {(-1, -1), (1, 1)}, vartype='SPIN')

        # get a penalty model with the desired properties
        with self.assertWarns(DeprecationWarning):
            pm = get_penalty_model(spec)

        with self.assertWarns(DeprecationWarning):
            pm2 = get_penalty_model(spec)

        self.assertEqual(pm, pm2)


class TestIterFactories(unittest.TestCase):
    def test_iter_factories(self):
        with self.assertWarns(DeprecationWarning):
            factory, = iter_factories()

        self.assertIs(factory, get_penalty_model)


class TestCoreNamespace(unittest.TestCase):
    def test_import(self):
        from penaltymodel.core import ImpossiblePenaltyModel


# Copyright 2022 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import itertools
import unittest
import unittest.mock

import dimod
import networkx as nx

from penaltymodel import get_penalty_model
from penaltymodel.database import isolated_cache


class TestGetPenaltyModel(unittest.TestCase):
    @isolated_cache()
    def test_different_energy_levels(self):
        samples_like = dimod.SampleSet.from_samples([[-1, -1, -1], [1, 1, 1]], energy=[0, .5], vartype='BINARY')

        bqm, gap = get_penalty_model(samples_like)

        self.assertAlmostEqual(bqm.energy([-1, -1, -1]), 0)
        self.assertAlmostEqual(bqm.energy([1, 1, 1]), .5)

    @isolated_cache()
    def test_different_energy_levels_graph_like(self):
        samples_like = dimod.SampleSet.from_samples([[-1, -1, -1], [1, 1, 1]], energy=[0, .5], vartype='BINARY')

        bqm, gap = get_penalty_model(samples_like, graph_like=nx.complete_graph(3))

        self.assertAlmostEqual(bqm.energy([-1, -1, -1]), 0)
        self.assertAlmostEqual(bqm.energy([1, 1, 1]), .5)

    @isolated_cache()
    def test_single_labelled(self):
        bqm, gap = get_penalty_model({'a': 1, 'b': 0})

        self.assertEqual(bqm, dimod.BQM({'a': -2.0, 'b': 2.0}, {('b', 'a'): 1.0}, 5.0, 'SPIN'))
        self.assertEqual(gap, 6)

        # now do it again, but make sure we use the cache
        with unittest.mock.patch('penaltymodel.interface.generate') as mock:
            mock.side_effect = Exception('boom')
            new = get_penalty_model({'a': 1, 'b': 0})

        self.assertEqual((bqm, gap), new)

    @isolated_cache()
    def test_subgraph_labelled(self):
        G = nx.Graph(itertools.product('abc', 'def'))

        bqm, gap = get_penalty_model(([[0, 0, 0], [0, 1, 0], [1, 0, 0], [1, 1, 1]], 'dbf'), G)

        ground = dimod.keep_variables(dimod.ExactSolver().sample(bqm), 'dbf').lowest().aggregate()

        self.assertEqual(len(ground), 4)
        for sample in ground.samples():
            self.assertEqual(sample['d'] > 0 and sample['b'] > 0, sample['f'] > 0)

    @isolated_cache()
    def test_unorded_range_labels(self):
        # NAE
        samples_like = ([(0, 0, 1, 0),
                         (0, 1, 1, 1),
                         (1, 0, 1, 1),
                         (0, 1, 0, 0),
                         (1, 0, 0, 0),
                         (1, 1, 0, 1)],
                        (0, 1, 3, 2))

        bqm, gap = get_penalty_model(samples_like)

        ground = dimod.ExactSolver().sample(bqm).lowest().aggregate()

        self.assertEqual(len(ground), 6)
        for sample in ground.samples():
            self.assertTrue(len(set(sample.values())) > 1)



# Copyright 2021 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest

import dimod
import networkx as nx
import numpy as np

from penaltymodel import MissingPenaltyModel
from penaltymodel.database import patch_cache


class TestBQMCache(unittest.TestCase):
    @patch_cache()
    def test_bqm_insert_retrieve(self, cache):
        bqm0 = dimod.generators.gnp_random_bqm(10, .5, 'SPIN', random_state=53)

        cache.insert_binary_quadratic_model(bqm0)
        self.assertEqual(list(cache.iter_binary_quadratic_models()), [bqm0])

        # insert again, should not duplicate
        cache.insert_binary_quadratic_model(bqm0)
        self.assertEqual(list(cache.iter_binary_quadratic_models()), [bqm0])

        # make a new bqm with reversed nodeorder
        bqm2 = dimod.BinaryQuadraticModel('SPIN')
        bqm2.add_linear_from((v, bqm0.get_linear(v)) for v in reversed(range(bqm0.num_variables)))
        bqm2.add_quadratic_from(bqm0.iter_quadratic())
        bqm2.offset = bqm0.offset

        cache.insert_binary_quadratic_model(bqm2)
        self.assertEqual(list(cache.iter_binary_quadratic_models()), [bqm0])

        bqm3 = dimod.BinaryQuadraticModel(bqm0, dtype=object)
        cache.insert_binary_quadratic_model(bqm3)
        self.assertEqual(list(cache.iter_binary_quadratic_models()), [bqm0])


class TestGraphCache(unittest.TestCase):
    def assertEqual(self, first, second, *args, **kwargs):
        if isinstance(first, nx.Graph) and isinstance(second, nx.Graph):
            self.assertEqual(first.nodes, second.nodes, *args, **kwargs)
            self.assertEqual(first.edges, second.edges, *args, **kwargs)
        else:
            return super().assertEqual(first, second, *args, **kwargs)

    @patch_cache()
    def test_graph_insert_retrieve(self, cache):
        graph = nx.barbell_graph(8, 8)
        nodelist = sorted(graph)
        edgelist = sorted(sorted(edge) for edge in graph.edges)

        cache.insert_graph(graph)

        self.assertEqual(list(cache.iter_graphs())[0], graph)

        # insert it again, and check that there is still only one graph
        cache.insert_graph(graph)
        self.assertEqual(len(list(cache.iter_graphs())), 1)

        # now adding another graph should result in two items
        graph = nx.complete_graph(4)
        cache.insert_graph(graph)
        self.assertEqual(len(list(cache.iter_graphs())), 2)

    @patch_cache()
    def test_graph_like(self, cache):
        cache.insert_graph(10)
        cache.insert_graph(range(10))
        cache.insert_graph(nx.complete_graph(10))
        cache.insert_graph(range(11))
        cache.insert_graph(nx.complete_graph(12))

        # we should see three graphs
        graphs = list(cache.iter_graphs())

        self.assertEqual(len(graphs), 3)
        self.assertTrue(any(nx.complete_graph(10).nodes == G.nodes for G in graphs))
        self.assertTrue(any(nx.complete_graph(11).nodes == G.nodes for G in graphs))
        self.assertTrue(any(nx.complete_graph(12).nodes == G.nodes for G in graphs))
        self.assertTrue(any(nx.complete_graph(10).edges == G.edges for G in graphs))
        self.assertTrue(any(nx.complete_graph(11).edges == G.edges for G in graphs))
        self.assertTrue(any(nx.complete_graph(12).edges == G.edges for G in graphs))


class TestPenaltyModelCache(unittest.TestCase):
    @patch_cache()
    def test_and_gate_insert_retrieve(self, cache):
        classical_gap = 2
        bqm = dimod.generators.and_gate(0, 1, 2, strength=classical_gap).change_vartype('SPIN', inplace=True)
        sampleset = dimod.ExactSolver().sample(bqm).lowest()

        cache.insert_penalty_model(bqm, sampleset, classical_gap)

        li = list(cache.iter_penalty_models())
        self.assertEqual(len(li), 1)
        pm, = li
        self.assertEqual(pm.bqm, bqm)
        data = list(pm.sampleset.data())
        for datum in sampleset.data():
            self.assertIn(datum, data)
        self.assertEqual(pm.classical_gap, classical_gap)


class TestRetrieve(unittest.TestCase):
    @patch_cache()
    def test_retrieve(self, cache):

        # put some AND gates into the cache
        bqm1 = dimod.generators.and_gate(0, 1, 2, strength=1).change_vartype('SPIN', inplace=True)
        cache.insert_penalty_model(bqm1, dimod.ExactSolver().sample(bqm1).lowest(), classical_gap=1)
        bqm2 = dimod.generators.and_gate(0, 1, 2, strength=2).change_vartype('SPIN', inplace=True)
        cache.insert_penalty_model(bqm2, dimod.ExactSolver().sample(bqm2).lowest(), classical_gap=2)

        samples = [[-1, -1, -1], [-1, +1, -1], [+1, -1, -1], [+1, +1, +1]]

        bqm, classical_gap = cache.retrieve(samples, nx.complete_graph(3))
        self.assertEqual(bqm, bqm2)  # largest gap, fits within bounds

        bqm, classical_gap = cache.retrieve(samples, nx.complete_graph(3), linear_bound=(-.5, .5), min_classical_gap=1)
        self.assertEqual(bqm, bqm1)  # since this fits in the given bounds

        with self.assertRaises(MissingPenaltyModel):
            cache.retrieve(samples, nx.complete_graph(3), linear_bound=(-.5, .5))


class TestSampleSetCache(unittest.TestCase):
    @patch_cache()
    def test_sampleset_insert_retrieve(self, cache):
        samples = [[-1, -1, +1], [+1, +1, +1]]
        cache.insert_sampleset(samples)
        sampleset, = cache.iter_samplesets()
        np.testing.assert_array_equal(samples, sampleset.record.sample)

    @patch_cache()
    def test_large(self, cache):
        samples = 2*np.random.randint(0, 2, size=(1, 32))-1
        cache.insert_sampleset(samples)
        sampleset, = cache.iter_samplesets()
        np.testing.assert_array_equal(samples, sampleset.record.sample)



# Copyright 2018 D-Wave Systems Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
import itertools

import networkx as nx

import penaltymodel.core as pm

import dimod


class TestSpecification(unittest.TestCase):
    def test_construction_empty(self):
        spec = pm.Specification(nx.Graph(), [], {}, dimod.SPIN)
        self.assertEqual(len(spec), 0)

    def test_construction_typical(self):
        graph = nx.complete_graph(10)
        decision_variables = (0, 4, 5)
        feasible_configurations = {(-1, -1, -1): 0.}

        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        self.assertEqual(spec.graph, graph)  # literally the same object
        self.assertEqual(spec.decision_variables, decision_variables)
        self.assertEqual(spec.feasible_configurations, feasible_configurations)
        self.assertIs(spec.vartype, dimod.SPIN)

    def test_construction_from_edgelist(self):
        graph = nx.barbell_graph(10, 7)
        decision_variables = (0, 4, 3)
        feasible_configurations = {(-1, -1, -1): 0.}

        # specification from edges
        spec0 = pm.Specification(graph.edges, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        # specification from graph
        spec1 = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        self.assertEqual(spec0, spec1)

    def test_construction_bad_graph(self):
        graph = 1
        decision_variables = (0, 4, 5)
        feasible_configurations = {(-1, -1, -1): 0.}

        with self.assertRaises(TypeError):
            pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

    def test_ranges_default(self):
        graph = nx.barbell_graph(4, 16)
        decision_variables = (0, 4, 3)
        feasible_configurations = {(0, 0, 0): 0.}

        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.BINARY)

        for v in graph:
            self.assertEqual(spec.ising_linear_ranges[v], [-2, 2])

        for u, v in graph.edges:
            self.assertEqual(spec.ising_quadratic_ranges[u][v], [-1, 1])
            self.assertEqual(spec.ising_quadratic_ranges[v][u], [-1, 1])

    def test_linear_ranges_specified(self):
        graph = nx.barbell_graph(4, 16)
        decision_variables = (0, 4, 3)
        feasible_configurations = {(0, 0, 1): 0.}

        spec = pm.Specification(graph, decision_variables, feasible_configurations,
                                ising_linear_ranges={v: [-v, 2] for v in graph},
                                vartype=dimod.BINARY)

        # check default energy ranges
        for v in graph:
            self.assertEqual(spec.ising_linear_ranges[v], [-v, 2])

        spec = pm.Specification(graph, decision_variables, feasible_configurations,
                                ising_linear_ranges={v: (-v, 2) for v in graph},
                                vartype=dimod.BINARY)

        # check default energy ranges
        for v in graph:
            self.assertEqual(spec.ising_linear_ranges[v], [-v, 2])

    def test_quadratic_ranges_partially_specified(self):
        graph = nx.barbell_graph(4, 16)
        decision_variables = (0, 4, 3)
        feasible_configurations = {(0, 0, 1): 0.}

        spec = pm.Specification(graph, decision_variables, feasible_configurations,
                                ising_quadratic_ranges={0: {1: [0, 1], 2: [-1, 0]}, 2: {0: [-1, 0]}},
                                vartype=dimod.BINARY)

        ising_quadratic_ranges = spec.ising_quadratic_ranges
        for u in ising_quadratic_ranges:
            for v in ising_quadratic_ranges[u]:
                self.assertIs(ising_quadratic_ranges[u][v], ising_quadratic_ranges[v][u])
        for u, v in graph.edges:
            self.assertIn(v, ising_quadratic_ranges[u])
            self.assertIn(u, ising_quadratic_ranges[v])

        self.assertEqual(ising_quadratic_ranges[0][1], [0, 1])

    def test_linear_ranges_bad(self):
        graph = nx.barbell_graph(4, 16)
        decision_variables = (0, 4, 3)
        feasible_configurations = {(0, 0, 1): 0.}

        with self.assertRaises(ValueError):
            pm.Specification(graph, decision_variables, feasible_configurations,
                             ising_linear_ranges={v: [-v, 'a'] for v in graph},
                             vartype=dimod.BINARY)

        with self.assertRaises(TypeError):
            pm.Specification(graph, decision_variables, feasible_configurations,
                             ising_linear_ranges={v: [-v, 1, 1] for v in graph},
                             vartype=dimod.BINARY)

    def test_vartype_specified(self):
        graph = nx.complete_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, 1): 0.}

        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)
        self.assertIs(spec.vartype, dimod.SPIN)

        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.BINARY)
        self.assertIs(spec.vartype, dimod.BINARY)

        # now set up a spec that can only have one vartype
        graph = nx.complete_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, -1): 0.}

        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)
        self.assertIs(spec.vartype, dimod.SPIN)

        # the feasible_configurations are spin
        with self.assertRaises(ValueError):
            spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.BINARY)

    def test_relabel_typical(self):
        graph = nx.circular_ladder_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, 1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        mapping = dict(enumerate('abcdefghijklmnopqrstuvwxyz'))

        new_spec = spec.relabel_variables(mapping, inplace=False)

        # create a test spec
        graph = nx.relabel_nodes(graph, mapping)
        decision_variables = (mapping[v] for v in decision_variables)
        test_spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        self.assertEqual(new_spec, test_spec)
        self.assertEqual(new_spec.ising_linear_ranges, test_spec.ising_linear_ranges)
        self.assertEqual(new_spec.ising_quadratic_ranges, test_spec.ising_quadratic_ranges)

    def test_relabel_copy(self):
        graph = nx.circular_ladder_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, 1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        mapping = dict(enumerate('abcdefghijklmnopqrstuvwxyz'))

        new_spec = spec.relabel_variables(mapping, inplace=False)

        # create a test spec
        graph = nx.relabel_nodes(graph, mapping)
        decision_variables = (mapping[v] for v in decision_variables)
        test_spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        self.assertEqual(new_spec, test_spec)
        self.assertEqual(new_spec.ising_linear_ranges, test_spec.ising_linear_ranges)
        self.assertEqual(new_spec.ising_quadratic_ranges, test_spec.ising_quadratic_ranges)

    def test_relabel_inplace(self):
        graph = nx.circular_ladder_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, 1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        mapping = {i: v for i, v in enumerate('abcdefghijklmnopqrstuvwxyz') if i in graph}

        new_spec = spec.relabel_variables(mapping, inplace=True)

        self.assertIs(new_spec, spec)  # should be the same object
        self.assertIs(new_spec.graph, spec.graph)

        # create a test spec
        graph = nx.relabel_nodes(graph, mapping)
        decision_variables = (mapping[v] for v in decision_variables)
        test_spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        self.assertEqual(new_spec, test_spec)
        self.assertEqual(new_spec.ising_linear_ranges, test_spec.ising_linear_ranges)
        self.assertEqual(new_spec.ising_quadratic_ranges, test_spec.ising_quadratic_ranges)

    def test_relabel_inplace_identity(self):
        graph = nx.circular_ladder_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, 1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        mapping = {v: v for v in graph}

        new_spec = spec.relabel_variables(mapping, inplace=True)

    def test_relabel_inplace_overlap(self):
        graph = nx.circular_ladder_graph(12)
        decision_variables = (0, 2, 5)
        feasible_configurations = {(1, 1, 1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        mapping = {v: v + 5 for v in graph}

        new_spec = spec.relabel_variables(mapping, inplace=True)

    def test_relabel_forwards_and_backwards(self):
        graph = nx.path_graph(4)
        graph.add_edge(0, 2)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        # make another one
        graph = nx.path_graph(4)
        graph.add_edge(0, 2)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        original_spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        identity = {v: v for v in graph}

        new_label_sets = [(10, 1),
                          ('a', 'b'),
                          (1, 'b'),
                          ('1', '2', '3', '4'),
                          ('a', 'b', 'c', 'd')]
        new_label_sets.extend(itertools.permutations(graph))

        for new in new_label_sets:
            mapping = dict(enumerate(new))
            inv_mapping = {u: v for v, u in mapping.items()}

            # apply then invert with inplace=False
            copy_spec = spec.relabel_variables(mapping, inplace=False)
            inv_copy = copy_spec.relabel_variables(inv_mapping, inplace=False)
            self.assertEqual(inv_copy, original_spec)
            self.assertEqual(inv_copy.ising_linear_ranges, original_spec.ising_linear_ranges)
            self.assertEqual(inv_copy.ising_quadratic_ranges, original_spec.ising_quadratic_ranges)

            # apply then invert with inplace=True
            spec.relabel_variables(mapping, inplace=True)
            if mapping == identity:
                self.assertEqual(spec, original_spec)
            else:
                self.assertNotEqual(spec, original_spec)
            spec.relabel_variables(inv_mapping, inplace=True)
            self.assertEqual(spec, original_spec)
            self.assertEqual(spec.ising_linear_ranges, original_spec.ising_linear_ranges)
            self.assertEqual(spec.ising_quadratic_ranges, original_spec.ising_quadratic_ranges)

    def test_bad_relabel(self):
        graph = nx.path_graph(4)
        graph.add_edge(0, 2)
        decision_variables = (0, 2)
        feasible_configurations = {(-1, -1): 0., (+1, +1): 0.}
        spec = pm.Specification(graph, decision_variables, feasible_configurations, vartype=dimod.SPIN)

        mapping = {0: 2, 1: 1}

        with self.assertRaises(ValueError):
            spec.relabel_variables(mapping, inplace=False)

        with self.assertRaises(ValueError):
            spec.relabel_variables(mapping, inplace=True)



# Copyright 2021 D-Wave Systems Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

# developer note: this combines all of the tests from maxgap and mip
# before we merged. There is likely a lot of redundancy

import itertools
import unittest

import dimod
import networkx as nx

from penaltymodel.generation import generate, ImpossiblePenaltyModel
from penaltymodel.utils import table_to_sampleset

MAX_GAP_DELTA = 0.01


class TestGenerate(unittest.TestCase):
    def check_bqm_table(self, bqm, gap, table, decision):
        """check that the bqm has ground states matching table"""
        response = dimod.ExactSolver().sample(bqm)

        highest_feasible_energy = max(table.values()) if isinstance(table, dict) else 0

        # Looping though ExactSolver results and comparing with BQM
        seen_gap = float('inf')
        seen_table = set()
        for sample, energy in response.data(['sample', 'energy']):
            self.assertAlmostEqual(bqm.energy(sample), energy)  # sanity check

            config = tuple(sample[v] for v in decision)

            # Configurations with energies < highest feasible energy
            if energy < highest_feasible_energy + .001:
                self.assertIn(config, table)

                # If there's no specific target energy to match, we've already
                # satisfied the configuration by being less than the highest
                # feasible energy
                if not isinstance(table, dict):
                    seen_table.add(config)
                    continue

                # Check configuration against specific target energy
                # Note: For a given valid decision configuration, there could
                #   be different sets of decision + auxiliary configurations. We
                #   only need one of those sets to match the specific target
                #   energy, while the remaining sets can be above that target.
                self.assertGreaterEqual(round(energy - table[config], ndigits=7), 0)

                # If configuration matches target energy, the configuration
                # should be added to the seen_table
                if round(table[config]-energy, ndigits=7) == 0:
                    seen_table.add(config)

            # Get smallest gap among non-table configurations
            elif config not in table:
                seen_gap = min(seen_gap, energy - highest_feasible_energy)

        # Verify that all table configurations have been accounted for
        for config in table:
            self.assertIn(config, seen_table)

        self.assertAlmostEqual(seen_gap, gap)
        self.assertGreater(round(gap, ndigits=7), 0)

    def check_bqm_graph(self, bqm, graph):
        """bqm and graph have the same structure"""
        self.assertEqual(len(bqm.linear), len(graph.nodes))
        self.assertEqual(len(bqm.quadratic), len(graph.edges))

        for v in bqm.linear:
            self.assertIn(v, graph)
        for u, v in bqm.quadratic:
            self.assertIn(u, graph.adj[v])

    def generate_and_check(self, graph, configurations, decision_variables,
                           *, known_classical_gap=0, **kwargs):

        bqm, gap, aux = generate(graph, table_to_sampleset(configurations, decision_variables), **kwargs)

        min_classical_gap = kwargs.get('min_classical_gap', 2)

        # Check gap
        # Note: Due to the way MaxGap searches for the maximum gap, if
        #   known_classical_gap == "maximum possible gap", then `gap` can be
        #   slightly smaller than known_classical_gap.
        self.assertGreaterEqual(round(gap, 9), min_classical_gap)
        self.assertGreaterEqual(gap, known_classical_gap - MAX_GAP_DELTA)

        # check that the bqm/graph have the same structure
        self.assertEqual(len(bqm.linear), len(graph.nodes))
        for v in bqm.linear:
            self.assertIn(v, graph.nodes)
        self.assertEqual(len(bqm.quadratic), len(graph.edges))
        for u, v in bqm.quadratic:
            self.assertIn((u, v), graph.edges)

        # now solve for the thing
        sampleset = dimod.ExactSolver().sample(bqm)

        # check that ground has 0 energy
        if len(sampleset):
            self.assertAlmostEqual(sampleset.first.energy, min(configurations.values()))

        # Get highest feasible energy
        if isinstance(configurations, dict) and configurations:
            highest_feasible_energy = max(configurations.values())
        else:
            highest_feasible_energy = 0

        # check gap and other energies
        best_gap = float('inf')
        seen = set()
        for sample, energy in sampleset.data(['sample', 'energy']):
            config = tuple(sample[v] for v in decision_variables)

            # we want the minimum energy for each config of the decision variables,
            # so once we've seen it once we can skip
            if config in seen:
                continue

            if config in configurations:
                self.assertAlmostEqual(energy, configurations[config])
                seen.add(config)
            else:
                best_gap = min(best_gap, energy - highest_feasible_energy)

        min_lin, max_lin = kwargs.get('linear_bound', (-2, 2))
        min_quad, max_quad = kwargs.get('quadratic_bound', (-1, 1))

        # check energy ranges
        for v, bias in bqm.linear.items():
            self.assertGreaterEqual(round(bias, 9), min_lin)
            self.assertLessEqual(round(bias, 9), max_lin)

        for (u, v), bias in bqm.quadratic.items():
            self.assertGreaterEqual(round(bias, 9), min_quad)
            self.assertLessEqual(round(bias, 9), max_quad)

        self.assertAlmostEqual(best_gap, gap)

    def test_disjoint(self):
        graph = nx.Graph()
        for u, v in itertools.product([0, 1, 2], [3, 4, 5]):
            graph.add_edge(u, v)
        graph.add_edge(8, 9)

        configurations = {(-1, -1, -1): 0,
                          (+1, +1, -1): 0}
        decision_variables = (0, 1, 8)

        self.generate_and_check(graph, configurations, decision_variables)

    def test_disjoint_decision_accross_subgraphs(self):
        graph = nx.Graph()
        for u, v in itertools.product([0, 1, 2], [3, 4, 5]):
            graph.add_edge(u, v)
        graph.add_edge(8, 9)

        configurations = {(-1, -1, +1, -1): 0,
                          (+1, +1, -1, -1): 0}
        decision_variables = (0, 1, 3, 8)

        self.generate_and_check(graph, configurations, decision_variables)

    def test_empty(self):
        # this should test things like empty graphs and empty configs
        graph = nx.Graph()
        configurations = {}
        decision_variables = tuple()

        self.generate_and_check(graph, configurations, decision_variables)

    def test_impossible(self):
        graph = nx.path_graph(3)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}
        decision_variables = (0, 1, 2)

        with self.assertRaises(ImpossiblePenaltyModel):
            generate(graph, table_to_sampleset(configurations, decision_variables))

    def test_K1(self):
        graph = nx.complete_graph(1)
        configurations = {(+1,): 0}
        decision_variables = [0]
        self.generate_and_check(graph, configurations, decision_variables)

    def test_K1_multiple_energies(self):
        graph = nx.complete_graph(1)
        configurations = {(+1,): .1, (-1,): -.3}
        decision_variables = [0]

        self.generate_and_check(graph, configurations, decision_variables)

    def test_K33(self):
        graph = nx.Graph()
        for i in range(3):
            for j in range(3, 6):
                graph.add_edge(i, j)

        decision_variables = (0, 2, 3)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}

        self.generate_and_check(graph, configurations, decision_variables)

    def test_K3_one_aux(self):
        graph = nx.complete_graph(3)
        configurations = {(-1, -1): 0, (1, 1): 0}
        decision_variables = [0, 1]

        self.generate_and_check(graph, configurations, decision_variables)

    def test_K4(self):
        graph = nx.complete_graph(4)

        configurations = {(-1, -1, -1, -1): 0, (1, 1, 1, 1): 0}
        decision_variables = list(graph)

        self.generate_and_check(graph, configurations, decision_variables)

    def test_min_gap_equals_max_gap(self):
        # Make sure that a model is always grabbed, even when min_gap == max_gap
        min_gap = 4     # This value is also the max classical gap
        decision_variables = ['a']
        config = {(-1,): -1}
        graph = nx.complete_graph(decision_variables)

        bqm, gap, aux = generate(graph, table_to_sampleset(config, decision_variables),
                                 min_classical_gap=min_gap)

        self.assertEqual(bqm, dimod.BinaryQuadraticModel({'a': 2}, {}, 1, 'SPIN'))
        self.assertEqual(min_gap, gap)

    def test_min_gap_no_aux(self):
        # Set up problem
        decision_variables = ['a', 'b', 'c']
        or_gate = {(-1, -1, -1): 0,
                   (-1, 1, 1): 0,
                   (1, -1, 1): 0,
                   (1, 1, 1): 0}
        graph = nx.complete_graph(decision_variables)

        # Run problem with a min_classical_gap that is too large
        with self.assertRaises(ImpossiblePenaltyModel):
            large_min_gap = 3
            generate(graph, table_to_sampleset(or_gate, decision_variables),
                     min_classical_gap=large_min_gap)

        # Lowering min_classical_gap should lead to a bqm being found
        min_classical_gap = 1.5
        self.generate_and_check(graph, or_gate, decision_variables,
                                min_classical_gap=min_classical_gap)

    def test_min_gap_with_aux(self):
        decision_variables = ['a', 'b', 'c']
        xor_gate = {(-1, -1, -1): 0,
                    (-1, 1, 1): 0,
                    (1, -1, 1): 0,
                    (1, 1, -1): 0}
        graph = nx.complete_graph(decision_variables + ['aux0'])

        # Run problem with a min_classical_gap that is too large
        with self.assertRaises(ImpossiblePenaltyModel):
            large_min_gap = 3
            generate(graph, table_to_sampleset(xor_gate, decision_variables),
                     min_classical_gap=large_min_gap)

        # Lowering min_classical_gap should lead to a bqm being found
        min_classical_gap = .5
        self.generate_and_check(graph, xor_gate, decision_variables,
                                min_classical_gap=min_classical_gap)

    def test_negative_min_gap_feasible_bqm(self):
        # Regardless of the negative min classical gap, this feasible BQM should return its usual
        # max classical gap.
        decision_variables = ['a']
        configurations = {(-1,): 0}
        graph = nx.complete_graph(decision_variables)
        min_classical_gap = -2

        self.generate_and_check(graph, configurations, decision_variables,
                                min_classical_gap=min_classical_gap)

    # def test_negative_min_gap_impossible_bqm(self):
    #     """XOR Gate problem without auxiliary variables
    #     Note: Regardless of the negative gap, this BQM should remain impossible.
    #     """
    #     negative_gap = -3
    #     decision_variables = ['a', 'b', 'c']
    #     xor_gate = {(-1, -1, -1): 0,
    #                 (-1, 1, 1): 0,
    #                 (1, -1, 1): 0,
    #                 (1, 1, -1): 0}
    #     graph = nx.complete_graph(decision_variables)

    #     with self.assertRaises(ImpossiblePenaltyModel):
    #         generate(graph, xor_gate, decision_variables,
    #                  min_classical_gap=negative_gap)

    def test_restricted_energy_ranges(self):
        """Create asymmetric energy ranges and test against that."""
        graph = nx.Graph()
        for u, v in itertools.product([0, 1, 2], [3, 4, 5]):
            graph.add_edge(u, v)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}
        decision_variables = (0, 1, 2)
        linear_bound = (-1., 2.)
        quadratic_bound = (-1, .5)

        self.generate_and_check(graph, configurations, decision_variables,
                                linear_bound=linear_bound,
                                quadratic_bound=quadratic_bound,
                                )

    def test_negative_feasible_positive_infeasible(self):
        """Testing that gap is wrt the energy of the highest feasible state, rather than wrt zero.

        Case where highest feasible state has negative energy and the gap is set high enough that
        the infeasible states must have positive energy.
        """
        min_classical_gap = 0.5
        decision_variables = ['a']
        configurations = {(1,): -0.5}
        graph = nx.complete_graph(decision_variables)

        # Known solution: -2*a + 1.5
        known_classical_gap = 4

        self.generate_and_check(graph, configurations, decision_variables,
                                known_classical_gap=known_classical_gap)

    def test_positive_feasible_positive_infeasible(self):
        """Testing that gap is wrt the energy of the highest feasible state, rather than wrt zero.

        Case where highest feasible state and the infeasible states must have positive energy.
        """
        min_classical_gap = 1
        decision_variables = ['a', 'b']
        configurations = {(1, -1): 4,
                          (-1, 1): 4,
                          (-1, -1): 0}
        graph = nx.complete_graph(decision_variables + ['c'])

        # Known solution: 2*a + 2*b -2*c + a*b + a*c + b*c + 7
        known_classical_gap = 8

        self.generate_and_check(graph, configurations, decision_variables,
                                min_classical_gap=min_classical_gap,
                                known_classical_gap=known_classical_gap)

    def test_negative_feasible_negative_infeasible(self):
        """Testing that gap is wrt the energy of the highest feasible state, rather than wrt zero.

        Case where highest feasible state and the infeasible states have negative energy levels.
        """
        min_classical_gap = 0.5
        decision_variables = ['a']
        configurations = {(1,): -10}
        graph = nx.complete_graph(decision_variables + ['b', 'c'])


        # Known solution: -2*a - 2*b - 2*c - a*b - a*c - b*c - 1
        known_classical_gap = 8

        self.generate_and_check(graph, configurations, decision_variables,
                                min_classical_gap=min_classical_gap,
                                known_classical_gap=known_classical_gap)

    def test_impossible_AND_3path(self):
        """AND gate cannot exist on a 3-path"""
        graph = nx.path_graph(3)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}
        decision_variables = (0, 1, 2)

        with self.assertRaises(ImpossiblePenaltyModel):
            generate(graph, table_to_sampleset(configurations, decision_variables))

    def test_empty_no_aux(self):
        graph = nx.Graph()
        sample = {}

        bqm, gap, aux = generate(graph, sample)

        self.check_bqm_graph(bqm, graph)

    def test_empty_some_aux(self):
        graph = nx.complete_graph(3)
        bqm, gap, aux = generate(graph, {})

        self.check_bqm_graph(bqm, graph)

    def test_AND_K4(self):
        """A typical use case, an AND gate on a K4."""
        graph = nx.complete_graph(4)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}
        decision_variables = (0, 1, 2)

        bqm, gap, aux = generate(graph, table_to_sampleset(configurations, decision_variables))

        self.check_bqm_table(bqm, gap, configurations, decision_variables)

    def test_NAE3SAT_4cycle(self):
        """A typical use case, an AND gate on a K4."""
        graph = nx.cycle_graph(4)
        configurations = {config: 0 for config in itertools.product((-1, 1), repeat=3) if len(set(config)) > 1}
        decision_variables = (0, 1, 2)

        bqm, gap, aux = generate(graph, table_to_sampleset(configurations, decision_variables))

        self.check_bqm_table(bqm, gap, configurations, decision_variables)

    def test_return_auxiliary_AND_K3(self):

        graph = nx.complete_graph(3)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}
        decision_variables = (0, 1, 2)

        bqm, gap, aux_configs = generate(graph, table_to_sampleset(configurations, decision_variables))

        # no aux variables
        self.assertEqual(aux_configs,
                         {(-1, -1, -1): {},
                          (-1, +1, -1): {},
                          (+1, -1, -1): {},
                          (+1, +1, +1): {}})

    def test_return_auxiliary_AND_K5(self):

        graph = nx.complete_graph(5)
        configurations = {(-1, -1, -1): 0,
                          (-1, +1, -1): 0,
                          (+1, -1, -1): 0,
                          (+1, +1, +1): 0}
        decision_variables = (0, 1, 2)

        bqm, gap, aux_configs = generate(graph, table_to_sampleset(configurations, decision_variables))

        for config in configurations:
            sample = dict(zip(decision_variables, config))

            sample.update(aux_configs[config])

            self.assertAlmostEqual(bqm.energy(sample), 0.0)
